@@ -1,31 +1,43 @@
 package Chemistry::OpenSMILES;
 
+# ABSTRACT: OpenSMILES format reader and writer
+our $VERSION = '0.10.0'; # VERSION
+
 use strict;
 use warnings;
 use 5.0100;
 
-# ABSTRACT: OpenSMILES format reader and writer
-our $VERSION = '0.8.4'; # VERSION
+use Chemistry::OpenSMILES::Stereo::Tables qw( @OH @TB );
+use Graph::Traversal::BFS;
+use List::Util qw( all any first none sum0 );
 
 require Exporter;
 our @ISA = qw( Exporter );
 our @EXPORT_OK = qw(
+    %bond_order_to_symbol
+    %bond_symbol_to_order
+    %normal_valence
     clean_chiral_centers
     is_aromatic
     is_chiral
+    is_chiral_allenal
+    is_chiral_octahedral
+    is_chiral_planar
+    is_chiral_tetrahedral
+    is_chiral_trigonal_bipyramidal
     is_cis_trans_bond
     is_double_bond
+    is_ring_atom
     is_ring_bond
     is_single_bond
+    is_triple_bond
     mirror
-    %normal_valence
     toggle_cistrans
+    valence
 );
 
-use Graph::Traversal::BFS;
-use List::Util qw(any);
-
 sub is_chiral($);
+sub is_chiral_planar($);
 sub is_chiral_tetrahedral($);
 sub mirror($);
 sub toggle_cistrans($);
@@ -44,12 +56,28 @@ our %normal_valence = (
     c  => [ 3 ], # Not from OpenSMILES specification
 );
 
-# Removes chiral setting from tetrahedral chiral centers with less than
-# four distinct neighbours. Only tetrahedral chiral centers with four atoms
-# are affected, thus three-atom centers (implying lone pairs) are left
-# untouched. Returns the affected atoms.
+our %bond_order_to_symbol = (
+    1   => '-',
+    1.5 => ':',
+    2   => '=',
+    3   => '#',
+    4   => '$',
+);
+
+our %bond_symbol_to_order = (
+    '-' => 1,
+    ':' => 1.5,
+    '=' => 2,
+    '#' => 3,
+    '$' => 4,
+);
+
+# Removes chiral setting from allenal, square planar, tetrahedral and trigonal bipyramidal chiral centers if deemed unimportant.
+# For allenal, tetrahedral and trigonal bipyramidal arrangements when not all the neighbours are distinct.
+# For square planar arrangements this means situations when all neighbours are the same.
+# Chiral centers with lone pairs are left untouched.
+# Returns the affected atoms.
 #
-# CAVEAT: disregards anomers
 # TODO: check other chiral centers
 sub clean_chiral_centers($$)
 {
@@ -57,18 +85,46 @@ sub clean_chiral_centers($$)
 
     my @affected;
     for my $atom ($moiety->vertices) {
-        next unless is_chiral_tetrahedral( $atom );
+        next unless is_chiral_allenal( $atom ) ||
+                    is_chiral_planar( $atom )  ||
+                    is_chiral_tetrahedral( $atom ) ||
+                    is_chiral_trigonal_bipyramidal( $atom );
+        # Anomers must not loose chirality settings in any way
+        next if is_ring_atom( $moiety, $atom, scalar $moiety->edges );
 
         my $hcount = exists $atom->{hcount} ? $atom->{hcount} : 0;
-        next if $moiety->degree($atom) + $hcount != 4;
+        my @neighbours = $moiety->neighbours( $atom );
+        if( is_chiral_allenal( $atom ) ) {
+            @neighbours = grep { $_ != $atom }
+                          map  { $moiety->neighbours( $_ ) }
+                               @neighbours;
+        }
+
+        if( is_chiral_trigonal_bipyramidal( $atom ) ) {
+            next if @neighbours + $hcount != 5;
+        } else {
+            next if @neighbours + $hcount != 4;
+        }
+
+        my %colors;
+        for (@neighbours, ( { symbol => 'H' } ) x $hcount) {
+            $colors{$color_sub->( $_ )}++;
+        }
+
+        if( is_chiral_planar( $atom ) ) {
+            # Chiral planar center markers make sense even if only two types of atoms are there.
+            next if scalar keys %colors  > 2;
+            next if scalar keys %colors == 2 && all { $_ == 2 } values %colors;
+        } elsif( is_chiral_trigonal_bipyramidal( $atom ) ) {
+            next if scalar keys %colors == 5;
+        } else {
+            next if scalar keys %colors == 4;
+        }
 
-        my %colors = map { ($color_sub->( $_ ) => 1) }
-                         $moiety->neighbours($atom),
-                         ( { symbol => 'H' } ) x $hcount;
-        next if scalar keys %colors == 4;
         delete $atom->{chirality};
         push @affected, $atom;
     }
+
     return @affected;
 }
 
@@ -88,17 +144,57 @@ sub is_chiral($)
     }
 }
 
+sub is_chiral_allenal($)
+{
+    my( $what ) = @_;
+    if( ref $what eq 'HASH' ) { # Single atom
+        return $what->{chirality} && $what->{chirality} =~ /^\@AL[12]$/;
+    } else {                    # Graph representing moiety
+        return any { is_chiral_allenal( $_ ) } $what->vertices;
+    }
+}
+
+sub is_chiral_planar($)
+{
+    my( $what ) = @_;
+    if( ref $what eq 'HASH' ) { # Single atom
+        return $what->{chirality} && $what->{chirality} =~ /^\@SP[123]$/;
+    } else {                    # Graph representing moiety
+        return any { is_chiral_planar( $_ ) } $what->vertices;
+    }
+}
+
 sub is_chiral_tetrahedral($)
 {
     my( $what ) = @_;
     if( ref $what eq 'HASH' ) { # Single atom
         # CAVEAT: will fail for allenal configurations of @/@@ in raw mode
-        return $what->{chirality} && $what->{chirality} =~ /^@@?$/
+        return $what->{chirality} && $what->{chirality} =~ /^@@?$/;
     } else {                    # Graph representing moiety
         return any { is_chiral_tetrahedral( $_ ) } $what->vertices;
     }
 }
 
+sub is_chiral_trigonal_bipyramidal($)
+{
+    my( $what ) = @_;
+    if( ref $what eq 'HASH' ) { # Single atom
+        return $what->{chirality} && $what->{chirality} =~ /^\@TB(1?[1-9]|20)$/;
+    } else {                    # Graph representing moiety
+        return any { is_chiral_trigonal_bipyramidal( $_ ) } $what->vertices;
+    }
+}
+
+sub is_chiral_octahedral($)
+{
+    my( $what ) = @_;
+    if( ref $what eq 'HASH' ) { # Single atom
+        return $what->{chirality} && $what->{chirality} =~ /^\@OH([1-9]|[12][0-9]|30)$/;
+    } else {                    # Graph representing moiety
+        return any { is_chiral_octahedral( $_ ) } $what->vertices;
+    }
+}
+
 sub is_cis_trans_bond
 {
     my( $moiety, $a, $b ) = @_;
@@ -113,18 +209,41 @@ sub is_double_bond
            $moiety->get_edge_attribute( $a, $b, 'bond' ) eq '=';
 }
 
+# An atom is deemed to be a ring atom if any of its bonds is a ring bond.
+sub is_ring_atom
+{
+    my( $moiety, $atom, $max_length ) = @_;
+    return '' unless $moiety->degree( $atom ) > 1;
+    return any { is_ring_bond( $moiety, $atom, $_, $max_length ) }
+               $moiety->neighbours( $atom );
+}
+
 # A bond is deemed to be a ring bond if there is an alternative path
 # joining its atoms not including the bond in consideration and this
 # alternative path is not longer than 7 bonds. This is based on
 # O'Boyle (2012) saying that Open Babel SMILES writer does not output
 # cis/trans markers for double bonds in rings of size 8 or less due to
 # them implicilty being cis bonds.
+#
+# If maximum ring size is given negative, ring size is not limited.
 sub is_ring_bond
 {
     my( $moiety, $a, $b, $max_length ) = @_;
-
     $max_length = 7 unless $max_length;
 
+    # A couple of shortcuts to reduce the complexity
+    return '' if any { $moiety->degree( $_ ) == 1 } ( $a, $b );
+    return '' if scalar( $moiety->vertices ) > scalar( $moiety->edges );
+
+    if( $max_length < 0 ) {
+        # Due to the issue in Graph, bridges() returns strings instead of real objects.
+        # Graph issue: https://github.com/graphviz-perl/Graph/issues/29
+        my %vertices_by_name = map { $_ => $_ } $moiety->vertices;
+        return none { ( $_->[0] == $a && $_->[1] == $b ) ||
+                      ( $_->[0] == $b && $_->[1] == $a ) }
+               map  { [ map { $vertices_by_name{$_} } @$_ ] } $moiety->bridges;
+    }
+
     my $copy = $moiety->copy;
     $copy->delete_edge( $a, $b );
 
@@ -133,7 +252,7 @@ sub is_ring_bond
         # Record number of bonds between $a and any other vertex
         my( $u, $v ) = @_;
         my @seen = grep { exists $distance{$_} } ( $u, $v );
-        return if @seen != 1; # Can this be 0?
+        return '' if @seen != 1; # Can this be 0?
 
         my $seen = shift @seen;
         my( $unseen ) = grep { !exists $distance{$_} } ( $u, $v );
@@ -160,14 +279,43 @@ sub is_single_bond
             $moiety->get_edge_attribute( $a, $b, 'bond' ) eq '-';
 }
 
+sub is_triple_bond
+{
+    my( $moiety, $a, $b ) = @_;
+    return $moiety->has_edge_attribute( $a, $b, 'bond' ) &&
+           $moiety->get_edge_attribute( $a, $b, 'bond' ) eq '#';
+}
+
 sub mirror($)
 {
     my( $what ) = @_;
     if( ref $what eq 'HASH' ) { # Single atom
-        # FIXME: currently dealing only with tetrahedral chiral centers
         if( is_chiral_tetrahedral( $what ) ) {
             $what->{chirality} = $what->{chirality} eq '@' ? '@@' : '@';
         }
+        if( is_chiral_allenal( $what ) ) {
+            $what->{chirality} = $what->{chirality} eq '@AL1' ? '@AL2' : '@AL1';
+        }
+        # Square planar centers are not affected by mirroring, doing nothing
+        if( is_chiral_trigonal_bipyramidal( $what ) ) {
+            my $number = substr $what->{chirality}, 3;
+            my $setting = $TB[$number-1];
+            my $opposite = first { $TB[$_]->{axis}[0] == $setting->{axis}[0] &&
+                                   $TB[$_]->{axis}[1] == $setting->{axis}[1] &&
+                                   $TB[$_]->{order}   ne $setting->{order} }
+                                 0..$#TB;
+            $what->{chirality} = '@TB' . ($opposite + 1);
+        }
+        if( is_chiral_octahedral( $what ) ) {
+            my $number = substr $what->{chirality}, 3;
+            my $setting = $OH[$number-1];
+            my $opposite = first { $OH[$_]->{shape}   eq $setting->{shape} &&
+                                   $OH[$_]->{axis}[0] == $setting->{axis}[0] &&
+                                   $OH[$_]->{axis}[1] == $setting->{axis}[1] &&
+                                   $OH[$_]->{order}   ne $setting->{order} }
+                                 0..$#OH;
+            $what->{chirality} = '@OH' . ($opposite + 1);
+        }
     } else {
         for ($what->vertices) {
             mirror( $_ );
@@ -180,6 +328,19 @@ sub toggle_cistrans($)
     return $_[0] eq '/' ? '\\' : '/';
 }
 
+sub valence($$)
+{
+    my( $moiety, $atom ) = @_;
+    return ($atom->{hcount} ? $atom->{hcount} : 0) +
+           sum0 map { exists $bond_symbol_to_order{$_}
+                           ? $bond_symbol_to_order{$_}
+                           : 1 }
+                map { $moiety->has_edge_attribute( $atom, $_, 'bond' )
+                           ? $moiety->get_edge_attribute( $atom, $_, 'bond' )
+                           : 1 }
+                    $moiety->neighbours( $atom );
+}
+
 # CAVEAT: requires output from non-raw parsing due issue similar to GH#2
 sub _validate($@)
 {
@@ -199,9 +360,8 @@ sub _validate($@)
             } elsif( $moiety->degree($atom) == 4 && $color_sub ) {
                 my %colors = map { ($color_sub->( $_ ) => 1) }
                                  $moiety->neighbours($atom);
-                if( scalar keys %colors != 4 ) {
-                    # FIXME: anomers are false-positives, see COD entry
-                    # 7111036
+                if( scalar keys %colors != 4 &&
+                    !is_ring_atom( $moiety, $atom, scalar $moiety->edges ) ) {
                     warn sprintf 'tetrahedral chiral setting for %s(%d) ' .
                                  'is not needed as not all 4 neighbours ' .
                                  'are distinct' . "\n",
@@ -405,7 +565,12 @@ convert neither implicit nor explicit hydrogen atoms in square brackets
 to atom hashes of their own. Moreover, it will not attempt to unify the
 representations of chirality. It should be noted, though, that many of
 subroutines of Chemistry::OpenSMILES expect non-raw data structures,
-thus processing raw output may produce distorted results.
+thus processing raw output may produce distorted results. In particular,
+C<write_SMILES()> calls from
+L<Chemistry::OpenSMILES::Writer|Chemistry::OpenSMILES::Writer> have to
+be instructed to expect raw data structure:
+
+    write_SMILES( \@moieties, { raw => 1 } );
 
 =back
 
